#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import deepcopy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Union
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.drop_duplicates_code_chunk import DropDuplicatesCodeChunk

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.transpiler.transpile_utils import (
    column_header_list_to_transpiled_code, column_header_to_transpiled_code)
from mitosheet.types import ColumnID

class DropDuplicatesStepPerformer(StepPerformer):
    """
    Allows you to drop duplicates from a dataframe
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'drop_duplicates'

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any]) -> Dict[str, Any]:
        return params

    @classmethod
    def execute( # type: ignore
        cls,
        prev_state: State,
        sheet_index: int,
        column_ids: List[ColumnID],
        keep: Union[str, bool],
        **params
    ):
        column_headers = [
            prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)
            for column_id in column_ids
        ]

        # If the subset is none, then we don't actually do the drop, as there are no
        # duplicates between 0 columns
        if len(column_headers) == 0:
            return None

        # We make a new state to modify it
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])

        pandas_start_time = perf_counter()
        final_df = post_state.dfs[sheet_index].drop_duplicates(
            subset=column_headers,
            keep=keep
        )
        pandas_processing_time = perf_counter() - pandas_start_time

        post_state.dfs[sheet_index] = final_df

        return post_state, {
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            DropDuplicatesCodeChunk(prev_state, post_state, params, execution_data)
        ]

    @classmethod
    def get_modified_dataframe_indexes( # type: ignore
        cls, 
        sheet_index: int,
        column_ids: List[ColumnID],
        keep: str,
        **params
    ) -> Set[int]:
        return {sheet_index}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import List

from mitosheet.code_chunks.code_chunk import CodeChunk


class DataframeDeleteCodeChunk(CodeChunk):

    def get_display_name(self) -> str:
        return 'Deleted Dataframe'
    
    def get_description_comment(self) -> str:
        sheet_index = self.get_param('sheet_index')
        df_name = self.prev_state.df_names[sheet_index]
        return f'Deleted {df_name}'

    def get_code(self) -> List[str]:
        old_dataframe_name = self.get_param('old_dataframe_name')

        return [f'del {old_dataframe_name}']


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import List, Optional
from mitosheet.code_chunks.code_chunk import CodeChunk


class EmptyCodeChunk(CodeChunk):
    """
    A code chunk that generates no code, but has a title. 

    NOTE: This is different than a NoOpCodeChunk, which we
    always want to be optimized out, as this has data we
    want to display to the user. Graphing steps are an example
    of this!
    """

    def get_display_name(self) -> str:
        return self.get_param('display_name')
    
    def get_description_comment(self) -> str:
        return self.get_param('description_comment')

    def get_code(self) -> List[str]:
        return []



from copy import copy
from typing import List
from mitosheet.code_chunks.code_chunk import CodeChunk


def optimize_code_chunks(all_code_chunks: List[CodeChunk]) -> List[CodeChunk]:
    """
    Given a list of code chunks, will recursively attempt to optimize them 
    down to the smallest list of code chunks that have the same effects
    as the original list. 

    This is necessarily recursive, because of a situation like [A, A, B, B], 
    where A and B can be combined to a No-op. Thus, after one call, we end with 
    [A, B], and we need to recurse to finish the optimization.
    """

    all_code_chunks_reversed = copy(all_code_chunks)
    all_code_chunks_reversed.reverse() # Reverse the list so we can pop from the old front quickly

    code_chunks_list: List[CodeChunk] = []

    optimized = False
    while len(all_code_chunks_reversed) >= 2:
        first_code_chunk = all_code_chunks_reversed.pop()
        second_code_chunk = all_code_chunks_reversed.pop()

        combined_chunk = first_code_chunk.combine_right(second_code_chunk)

        if combined_chunk is not None:
            # If we can combine the two chunks into one, take that
            optimized = True
            all_code_chunks_reversed.append(combined_chunk)
        else:
            # If we cannot combine the two chunks, take the first chunk, 
            # and reset the second one to the new first chunk for the next loop
            code_chunks_list.append(first_code_chunk)
            all_code_chunks_reversed.append(second_code_chunk)            
    
    # Make sure we take the final item in the code chunks list, 
    # as it has nothing to combine_right with
    if len(all_code_chunks_reversed) == 1:
        code_chunks_list.append(all_code_chunks_reversed[0])

    # TODO: we could combine_left here? This would allow us to 
    # express overwriting conditions very naturally for deleting
    # dataframes...

    # As long as we optimized in this iteration, recurse as we might
    # need to optimize again
    if optimized:
        return optimize_code_chunks(code_chunks_list)

    return code_chunks_list